import re

from docutils.core import publish_doctree
from recommonmark.parser import CommonMarkParser

from .models import Issue, Metadata

META_REGEX = re.compile(r'^:(state|assignee|labels|milestone):\s*?(.*)\r?\n?',
                        flags=re.MULTILINE)


def get_metadata(text):
    meta = Metadata((k, v.strip()) for k, v in re.findall(META_REGEX, text))
    if 'labels' in meta:
        meta['labels'] = [l.strip() for l in meta['labels'].split(',')]
    return meta


def clean_metadata(text):
    return re.sub(META_REGEX, '', text)


def parser(path):

    with open(path) as fh:
        source = fh.read()
    lines = source.split('\n')
    dt = publish_doctree(source, parser=CommonMarkParser())

    tokens = {}

    for i, (sec_id, section) in enumerate(dt.ids.items()):
        if section.parent.tagname != 'document':
            continue
        text = section.astext()
        title = text.split('\n')[0].strip()
        number = None
        matched = re.match(r'^\[#(\d+)]\s*?(.*)$', title)
        if matched:
            number, title = matched.groups()

        metadata = get_metadata(text)

        tokens[section.line] = (title.strip(),
                                int(number) if number else None,
                                metadata)
    k = list(sorted(tokens.keys()))
    k.append(len(lines) + 2)
    return [Issue(title=tokens[s][0],
                  number=tokens[s][1],
                  metadata=tokens[s][2],
                  body=clean_metadata('\n'.join(lines[s:e - 2]))) for s, e in zip(k, k[1:])]


if __name__ == '__main__':
    import sys
    parser(sys.argv[0])



from .models import Issue, Metadata
from github import Github, GithubException, GithubObject


class GithubManager(object):

    def __init__(self, repo, login_or_token=None, password=None):
        self.g = Github(login_or_token, password)
        self.repo = self.g.get_repo(repo)

    def fetcher(self, state):
        issues = []
        for issue in self.repo.get_issues(state=state):
            meta = Metadata()
            if issue.state != 'open':
                # assume state open
                meta['state'] = issue.state
            if issue.milestone:
                meta['milestone'] = issue.milestone.title
            if issue.labels:
                meta['labels'] = [l.name for l in issue.labels]
            if issue.assignee:
                meta['assignee'] = issue.assignee.login
            issues.append(Issue(title=issue.title,
                                body=issue.body,
                                number=issue.number,
                                metadata=meta))
        return issues

    def get_milestone(self, title):
        """
        given the title as str, looks for an existing milestone or create a new one,
        and return the object
        """
        if not title:
            return GithubObject.NotSet
        if not hasattr(self, '_milestones'):
            self._milestones = {m.title: m for m in self.repo.get_milestones()}

        milestone = self._milestones.get(title)
        if not milestone:
            milestone = self.repo.create_milestone(title=title)
        return milestone

    def get_assignee(self, login):
        """
        given the user login, looks for a user in assignee list of the repo
        and return it if was found.
        """
        if not login:
            return GithubObject.NotSet
        if not hasattr(self, '_assignees'):
            self._assignees = {c.login: c for c in self.repo.get_assignees()}
        if login not in self._assignees:
            # warning
            print("{} doesn't belong to this repo. This issue won't be assigned.".format(login))
        return self._assignees.get(login)

    def get_state(self, state):
        if not state:
            return 'open'
        if state not in ('open', 'closed'):
            print("{} isn't a valid state (i.e: open or closed)".format(state))
            return 'open'
        return state

    def sender(self, issues):
        """
        push a list of issues to github
        """

        for issue in issues:
            state = self.get_state(issue.state)
            if issue.number:
                try:
                    gh_issue = self.repo.get_issue(issue.number)
                    original_state = gh_issue.state
                    if original_state == state:
                        action = 'Updated'
                    elif original_state == 'closed':
                        action = 'Reopened'
                    else:
                        action = 'Closed'

                    gh_issue.edit(title=issue.title,
                                  body=issue.body,
                                  labels=issue.labels,
                                  milestone=self.get_milestone(issue.milestone),
                                  assignee=self.get_assignee(issue.assignee),
                                  state=self.get_state(issue.state)
                                  )
                    print('{} #{}: {}'.format(action, gh_issue.number, gh_issue.title))
                except GithubException:
                    print('Not found #{}: {} (ignored)'.format(issue.number, issue.title))
                    continue
            else:
                gh_issue = self.repo.create_issue(title=issue.title,
                                                  body=issue.body,
                                                  labels=issue.labels,
                                                  milestone=self.get_milestone(issue.milestone),
                                                  assignee=self.get_assignee(issue.assignee))
                print('Created #{}: {}'.format(gh_issue.number, gh_issue.title))








"""
Tisú: your issue tracker, in a text file

Usage:
  tisu push <markdown_file> [--repo=<repo>] [(--username=<username> [--pass=<pass>]|--token=<token>)]
  tisu pull <markdown_file> [--repo=<repo>] [--state=<state>] [(--username=<username> [--pass=<pass>]|--token=<token>)]

Options:
  -h --help                 Show this screen.
  --version                 Show version.
  --repo=<repo>             Github repo (as: user/name). [default: inferred from git remote]
  --state=<state>           Filter by issue state [default: open].
  --username=<username>     Github username to send issues. Repo's username if no given.
  --pass=<pass>             Github password. Prompt if user is given and it is not.
  --token=<token>           Personal app token. Default to GITHUB_TOKEN environment variable.
                            Get one at https://github.com/settings/tokens
"""
import os
import re
from getpass import getpass
from subprocess import check_output

from docopt import docopt

from .gh import GithubManager
from .parser import parser

__version__ = "2.1"


def pull(repo, path, state, username_or_token=None, password=None):
    issues = GithubManager(repo, username_or_token, password).fetcher(state)
    with open(path, 'w') as fh:
        for issue in issues:
            fh.write(str(issue))


def push(path, repo, username_or_token, password):
    issues = parser(path)
    issues = GithubManager(repo, username_or_token, password).sender(issues)


def github_from_git():
    s = check_output(['git', 'remote', '-v'])
    return re.findall(r'[\w\-]+\/[\w\-]+', s.decode('utf8'))[0]


def main():
    args = docopt(__doc__, version=__version__)
    repo = args['--repo'] if args['--repo'] != 'inferred from git remote' else github_from_git()
    token = args.get('--token') or os.environ.get("GITHUB_TOKEN")
    
    username = args.get('--username', repo.split('/')[0])
    password = args.get('--pass', getpass('Github password: ') if not token else None)

    if args['pull']:
        pull(repo, args['<markdown_file>'], args['--state'], username_or_token=token or username, password=password)
    elif args['push']:
        push(args['<markdown_file>'], repo, token or username, password)


if __name__ == '__main__':
    main()



import os.path
from tisu.parser import parser, get_metadata, clean_metadata


def s(file):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'sources', file)


def test_multiples_issues():
    issues = parser(s('issues.md'))
    assert len(issues) == 3
    assert issues[0].title == 'Fix the parser'
    assert issues[0].body == '\nThis would be the content of and issue\n\n## subseccion 1.1\n\nthis is a subsection, part of the issue body\n'
    assert issues[1].title == 'Improve tests'
    assert issues[1].body == '\nAnother issue'
    assert issues[2].title == 'Be happy'
    assert issues[2].body == "\nThe life's milestone.\n"


def test_with_number():
    issues = parser(s('issue_with_number.md'))
    assert len(issues) == 1
    assert issues[0].title == 'Fix the parser'
    assert issues[0].number == 18


def test_get_metadata():
    meta = get_metadata(open(s('with_metadata.md')).read())
    assert meta == {'assignee': 'mgaitan',
                    'labels': ['x', 'y', 'z'],
                    'milestone': 'sprint1',
                    'state': 'open'}


def test_get_metadata_is_stable():
    meta = get_metadata(open(s('with_metadata.md')).read())
    assert meta == get_metadata(str(meta))


def test_no_metadata_return_no_output():
    meta = get_metadata('No\n\nmetadata')
    assert str(meta) == ''


def test_clean_metadata():
    text = clean_metadata(open(s('with_metadata.md')).read())
    assert text == '# test1\n\n\nbody\n'


def test_clean_metadata_not_a_block():
    text = clean_metadata("""# title

:labels: x,b,z
some content
:milestone:sprint1

more content
""")
    assert text == '# title\n\nsome content\n\nmore content\n'


def test_with_metatada():
    issues = parser(s('with_metadata.md'))
    assert len(issues) == 1
    assert issues[0].title == 'test1'
    assert issues[0].number is None
    assert issues[0].body == '\n\nbody\n'
    assert issues[0].metadata == {'assignee': 'mgaitan',
                                  'labels': ['x', 'y', 'z'],
                                  'milestone': 'sprint1'}




class Metadata(dict):

    def __str__(self):
        metadata = []
        for k, v in self.items():
            if isinstance(v, list):
                v = ', '.join(v)
            metadata.append(":{}: {}".format(k, v))
        return '\n'.join(metadata)


class Issue(object):
    def __init__(self, title, body, number=None, metadata=None, *kwargs):
        self.title = title
        self.body = body
        self.number = number
        self.metadata = metadata if metadata else Metadata()

    def __repr__(self):
        if self.number:
            return "<{0.__class__.__name__}: [#{0.number}] {0.title}>".format(self)
        return "<{0.__class__.__name__}: {0.title}>".format(self)

    def __str__(self):

        s = '\n\n'.join((v.strip() for v in (self.title,
                                             str(self.metadata),
                                             self.body) if v and v.strip()))
        if self.number:
            return "# [#{}] {}\n\n".format(self.number, s)
        return "# {}\n\n".format(s)

    @property
    def labels(self):
        return self.metadata.get('labels') or []

    @property
    def milestone(self):
        return self.metadata.get('milestone')

    @property
    def assignee(self):
        return self.metadata.get('assignee')

    @property
    def state(self):
        return self.metadata.get('state')
